/* LibTomCrypt, modular cryptographic library -- Tom St Denis
 *
 * LibTomCrypt is a library that provides various cryptographic
 * algorithms in a highly modular and flexible manner.
 *
 * The library is free for all purposes without any express
 * guarantee it works.
 *
 * Tom St Denis, tomstdenis@gmail.com, http://libtom.org
 */

/**
   @file gcm_done.c
   GCM implementation, Terminate the stream, by Tom St Denis
*/
#include "tomcrypt.h"

#ifdef LTC_GCM_MODE

/**
  Terminate a GCM stream
  @param gcm     The GCM state
  @param tag     [out] The destination for the MAC tag
  @param taglen  [in/out]  The length of the MAC tag
  @return CRYPT_OK on success
 */
int gcm_done(gcm_state *gcm, 
                     unsigned char *tag,    unsigned long *taglen)
{
   unsigned long x;
   int err;

   LTC_ARGCHK(gcm     != NULL);
   LTC_ARGCHK(tag     != NULL);
   LTC_ARGCHK(taglen  != NULL);

   if (gcm->buflen > 16 || gcm->buflen < 0) {
      return CRYPT_INVALID_ARG;
   }

   if ((err = cipher_is_valid(gcm->cipher)) != CRYPT_OK) {
      return err;
   }


   if (gcm->mode != LTC_GCM_MODE_TEXT) {
      return CRYPT_INVALID_ARG;
   }
#ifdef LTC_GCM_TABLES
fprintf(stderr, "DEBUG from gcm_done.c (LTC_GCM_TABLES defined):\n");
#else
fprintf(stderr, "DEBUG from gcm_done.c (LTC_GCM_TABLES NOT defined):\n");
#endif
fprintf(stderr, "gcm_done:X_0="); for(x=0;x<16;x++) fprintf(stderr, "%02x", gcm->X[x]); fprintf(stderr, "\n");
   /* handle remaining ciphertext */
   if (gcm->buflen) {
      gcm->pttotlen += gcm->buflen * CONST64(8);
      gcm_mult_h(gcm, gcm->X);
   }
fprintf(stderr, "gcm_done:X_1="); for(x=0;x<16;x++) fprintf(stderr, "%02x", gcm->X[x]); fprintf(stderr, "\n");
   /* length */
   STORE64H(gcm->totlen, gcm->buf);
   STORE64H(gcm->pttotlen, gcm->buf+8);
fprintf(stderr, "gcm_done:BUF="); for(x=0;x<16;x++) fprintf(stderr, "%02x", gcm->buf[x]); fprintf(stderr, "\n");
   for (x = 0; x < 16; x++) {
       gcm->X[x] ^= gcm->buf[x];
   }
fprintf(stderr, "gcm_done:H_0="); for(x=0;x<16;x++) fprintf(stderr, "%02x", gcm->H[x]); fprintf(stderr, "\n");
fprintf(stderr, "gcm_done:X_2="); for(x=0;x<16;x++) fprintf(stderr, "%02x", gcm->X[x]); fprintf(stderr, "\n");
   gcm_mult_h(gcm, gcm->X);
fprintf(stderr, "gcm_done:X_3="); for(x=0;x<16;x++) fprintf(stderr, "%02x", gcm->X[x]); fprintf(stderr, "\n");
fprintf(stderr, "gcm_done:Y_0="); for(x=0;x<16;x++) fprintf(stderr, "%02x", gcm->Y_0[x]); fprintf(stderr, "\n");
   /* encrypt original counter */
   if ((err = cipher_descriptor[gcm->cipher].ecb_encrypt(gcm->Y_0, gcm->buf, &gcm->K)) != CRYPT_OK) {
      return err;
   }
fprintf(stderr, "gcm_done:X_4="); for(x=0;x<16;x++) fprintf(stderr, "%02x", gcm->X[x]); fprintf(stderr, "\n");
fprintf(stderr, "gcm_done:BUF="); for(x=0;x<16;x++) fprintf(stderr, "%02x", gcm->buf[x]); fprintf(stderr, "\n");
   for (x = 0; x < 16 && x < *taglen; x++) {
       tag[x] = gcm->buf[x] ^ gcm->X[x];
   }
   *taglen = x;
fprintf(stderr, "gcm_done:TAG="); for(x=0;x<*taglen;x++) fprintf(stderr, "%02x", tag[x]); fprintf(stderr, "\n");

   cipher_descriptor[gcm->cipher].done(&gcm->K);

   return CRYPT_OK;
}

#endif


/* $Source$ */
/* $Revision$ */
/* $Date$ */
